%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% JUEGO DEL ZORRO Y LOS SABUESOS          %
% Autores: Jaime Rascón | Anghel Dulceanu %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Los nombres de variables suelen ser:
% XZ = Posición X del Zorro
% YZ = Posición Y del Zorro
% LS = Lista de Sabuesos
% NLS = Nueva Lista de Sabuesos
% D = Dirección
% S = Número de Sabueso

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% REPRESENTACION DEL JUEGO %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% TABLERO

% [0,s,0,s,0,s,0,s]
% [0,0,0,0,0,0,0,0]
% [0,0,0,0,0,0,0,0]
% [0,0,0,0,0,0,0,0]
% [0,0,0,0,0,0,0,0]
% [0,0,0,0,0,0,0,0]
% [0,0,0,0,0,0,0,0]
% [z,0,0,0,0,0,0,0]

% Ponemos las posiciones de 0 a longitud-1, como con arrays
posicionZorro(7,0).
posicionSabuesos([[0,1],[0,3],[0,5],[0,7]]).

%%%%%%%%%%%%%%
% CONSTANTES %
%%%%%%%%%%%%%%

% Direcciones: 1,2,3,4 para el zorro, 3,4 para los sabuesos
% Tablero de 8x8

%%%%%%%
% 1 2 %
%  Z  %
% 3 4 %
%%%%%%%

longitudTablero(8).
direccion(D) :- member(D,[1,2,3,4]).
direccionSabueso(D) :- member(D,[3,4]).
numeroSabueso(S) :- member(S,[0,1,2,3]).
columnasPosibles(X,Y):-par(X)-> (member(Y,[1,3,5,7]));(member(Y,[0,2,4,6])).
fila(F) :- member(F,[0,1,2,3,4,5,6,7]).

%%%%%%%%%%%%%%%%%%
% INICIALIZACION %
%%%%%%%%%%%%%%%%%%		
:- use_module(library(clpfd)).
:- initialization(nuevaPartida).

%%%%%%%%%%%%
% GRAFICOS %
%%%%%%%%%%%%

% Limpiar la pantalla
limpiar:-format('~c~s', [0x1b, "[2J"]).

% ---------- PINTAR MAPA A PARTIR DE LAS POSICIONES DE FICHAS ---------- %
% Parámetros: LZ, lista de zorros, LS, lista de sabuesos
pintarTablero(XZ,YZ,LS) :- 
	nl,
	writeln('================'),
	writeln('ZORRO Y SABUESOS'),
	writeln('================'),
	nl,
	pintar(0,0,XZ,YZ,LS).


pintar(X,Y,XZ,YZ,LS) :- longitudTablero(Long), (
	Y#<Long-1, pintarCasilla(X,Y,XZ,YZ,LS), Y1 is Y+1, pintar(X,Y1,XZ,YZ,LS), !;
	X#<Long-1, Y#=Long-1, pintarCasilla(X,Y,XZ,YZ,LS), X1 is X+1, nl, pintar(X1,0,XZ,YZ,LS), !;
	X#=Long-1, Y#=Long-1, pintarCasilla(X,Y,XZ,YZ,LS), nl
	).

% Pinta lo que haya en una casilla determinada	
pintarCasilla(X,Y,XZ,YZ,LS) :- pintarZorro(X,Y,XZ,YZ),! ; pintarSabueso(X,Y,LS),! ; pintarVacio.

% Comprueba si en una casilla hay un sabueso, en cuyo caso lo pinta
pintarSabueso(X,Y,LS) :- member([X,Y],LS), ansi_format([bold,fg(red)], '~w', [s]), write(' ').

% Comprueba si en una casilla hay un zorro, en cuyo caso lo pinta
pintarZorro(X,Y,X,Y) :- ansi_format([bold,fg(blue)], '~w', [z]) ,write(' ').
pintarVacio :- ansi_format([fg(black)], '~w', [o]), write(' ').	

%%%%%%%%%%%%%%%%%%%%
% LOGICA DEL JUEGO %
%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%
% Mover las fichas %
%%%%%%%%%%%%%%%%%%%%

% Mueve al zorro en dirección diagonal D
moverZorro(D,XZ,YZ,LS,NXZ,NYZ) :- posSgte(D,XZ,YZ,NXZ,NYZ), posicionVacia(NXZ,NYZ,XZ,YZ,LS).	

% Mueve al sabueso número S en dirección diagonal D
moverSabueso(S,D,XZ,YZ,LS,NL) :- dameSabueso(S,LS,[X,Y]),
	direccionSabueso(D),
	posSgte(D,X,Y,NX,NY),
	posicionVacia(NX,NY,XZ,YZ,LS),
	cambiarLista(S,[NX,NY],LS,NL)
	.

% Si el zorro puede moverse en alguna dirección
puedeMoverZorro(XZ,YZ,LS) :- direccion(D), puedeMoverZorro(D,XZ,YZ,LS).

% Si el zorro puede moverse en una dirección concreta
puedeMoverZorro(D,XZ,YZ,LS):-
	posSgte(D,XZ,YZ,NX,NY),
	posicionVacia(NX,NY,XZ,YZ,LS).

% Si la posicion se puede ocupar
posicionVacia(X,Y,XZ,YZ,LS) :- posicionPosible(X,Y), posicionLibre(X,Y,XZ,YZ,LS).

% Si la posicion no se sale del tablero
posicionPosible(X,Y) :- longitudTablero(Long), X>=0, Y>=0, X=<Long-1, Y=<Long-1.

% Si la posicion no está ocupada
posicionLibre(X,Y,XZ,YZ,LS) :- posicionSinZorro(X,Y,XZ,YZ), posicionSinSabuesos(X,Y,LS).

% Si la posicion no está ocupada por un zorro
posicionSinZorro(X,Y,XZ,YZ) :- X =\= XZ; Y =\= YZ.

% Si la posicion no está ocupada por un sabueso
posicionSinSabuesos(X,Y,L) :- not(member([X,Y],L)).

% Simula una asignación en la posición S de un array (lista[S] = [X,Y])
cambiarLista(S,X,[_|Zs],[X|Zs]):- S==0, !.
cambiarLista(S,X,[Z|Zs],[Z|Vs]):- S>0, S1 is S-1,cambiarLista(S1,X,Zs,Vs). 

% Devuelve la posición del sabueso número "S" en "X" (X = lista[S])
dameSabueso(0,[X|_],X) :- !.
dameSabueso(S,[_|Xs],Y) :- S>0, S1 is S-1, dameSabueso(S1,Xs,Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CONDICIONES DE VICTORIA %
%%%%%%%%%%%%%%%%%%%%%%%%%%%

finJuego(XZ,YZ,LS) :- ganaSabuesos(XZ,YZ,LS), !; ganaZorro(XZ,YZ,LS), !.

ganaSabuesos(XZ,YZ,LS) :- XZ=\=0, not(puedeMoverZorro(XZ,YZ,LS)), !, limpiar, pintarTablero(XZ,YZ,LS), nl, ansi_format([bold,fg(black)], 'Han ganado los ~w', [sabuesos]), nl, nl.
ganaZorro(XZ,YZ,LS) :- XZ==0, !, limpiar, pintarTablero(XZ,YZ,LS), nl, ansi_format([bold,fg(black)], 'Ha ganado el ~w', [zorro]), nl, nl.

% FALTA IMPLEMENTAR ESTE CASO -> puedeMoverZorro, not(puedenMoverSabuesos).
% puedeGanarZorro(XZ,YZ,LS) :- puedeMoverZorro(XZ,YZ,LS), not(puedenMoverSabuesos(XZ,YZ,LS)) , writeln('El zorro puede mover y los sabuesos no').

%%%%%%%%%%%%%%%%%%%
% INICIAR PARTIDA %
%%%%%%%%%%%%%%%%%%%

nuevaPartida :- posicionZorro(XZ,YZ), posicionSabuesos(LS),
	nl,
	writeln('================'),
	writeln('ZORRO Y SABUESOS'),
	writeln('================'),
	nl,
	writeln('Esta es una implementacion en prolog del juego de los sabuesos'),
	writeln('Para mover al zorro debes insertar la direccion del movimiento de 1 a 4'),
	writeln('El zorro solo puede moverse a una casilla libre en diagonal, se representa con una Z'),
	writeln('Los sabuesos se representan con una S'),
	nl,
	writeln('1 2'),
	ansi_format([bold,fg(blue)], ' ~w', [z]), nl,
	writeln('3 4'),
	nl,	
	escribeMenu(XZ,YZ,LS).
	
escribeMenu(XZ,YZ,LS) :- 
	writeln('=[ OPCIONES DE JUEGO ]='),
	nl,
	writeln('1. Jugador vs Jugador'),
	writeln('2. Jugador vs PC (Dificultad: Facil)'),
	writeln('3. Jugador vs PC (Dificultad: Dificil)'),
	nl,
	writeln('Introduzca "1.", "2." o "3." para seleccionar la opcion y pulse ENTER'),
	read(O),
	opcionJuego(O,XZ,YZ,LS).

opcionJuego(O,XZ,YZ,LS) :- ( 
	(O == 1) -> 
	(
		jugar1VS1(XZ,YZ,LS)
	)
	;  
	(O == 2)->
	(
		writeln('Ha elegido Jugador vs PC (Facil)'),sleep(1),
		jugarCPUFacil(XZ,YZ,LS)
	)
	;
	(O == 3)->
	(
		writeln('Ha elegido Jugador vs PC (Dificil)'),sleep(1),
		jugarCPUDificil(XZ,YZ,LS)
	)
	;
	(O =\= 1, O=\=2, O=\=3)->
	(
		limpiar,
		writeln('Valor incorrecto; Vuelva a introducir la opcion!'),
		escribeMenu(XZ,YZ,LS)
	)
).

%%%%%%%%%%%%%%%%%%%%
%  BUCLES DE JUEGO %
%%%%%%%%%%%%%%%%%%%%

% Ejecuta un juego contra otro jugador
% Comprueba condicion de victoria después de que mueva cada uno
% Así si el zorro llega al final, gana y no se le obliga al sabueso a mover
jugar1VS1(XZ,YZ,LS) :- 
	turnoZorro(XZ,YZ,NXZ,NYZ,LS),
	(
		finJuego(NXZ,NYZ,LS), !;
		(
			turnoSabueso(NXZ,NYZ,LS,NLS),
			(
				finJuego(NXZ,NYZ,NLS), !;
				jugar1VS1(NXZ,NYZ,NLS)
			)
		)
	)
	.

% Ejecuta un juego contra la CPU
jugarCPUFacil(XZ,YZ,LS) :- 
	turnoZorro(XZ,YZ,NXZ,NYZ,LS),
	turnoSabuesoCPUFacil(NXZ,NYZ,LS,NLS),
	(finJuego(NXZ,NYZ,NLS) ,! ; jugarCPUFacil(NXZ,NYZ,NLS))
	.

% Ejecuta un juego contra la CPU
jugarCPUDificil(XZ,YZ,LS) :- 
	turnoZorro(XZ,YZ,NXZ,NYZ,LS),
	turnoSabuesoCPUDificil(NXZ,NYZ,LS,NLS),
	(finJuego(NXZ,NYZ,NLS) ,! ; jugarCPUDificil(NXZ,NYZ,NLS))
	.

% Ejecuta el turno del zorro
turnoZorro(XZ,YZ,NXZ,NYZ,LS) :- 
	limpiar,
	pintarTablero(XZ,YZ,LS),
	nl,
	writeln('Te toca mover al zorro: Direcciones -> 1,2,3,4'),
	nl,
	read(D),
	(moverZorro(D,XZ,YZ,LS,NXZ,NYZ) -> 
		(
			% No hacer nada
			true
		)
		;
		(
			writeln('El movimiento no se puede realizar, elija otro movimiento.'),sleep(1),
			turnoZorro(XZ,YZ,NXZ,NYZ,LS)
		)
	).

% Ejecuta el turno de los sabuesos (En partida 1vs1)
turnoSabueso(XZ,YZ,LS,NLS) :- 
	limpiar,
	pintarTablero(XZ,YZ,LS),
	nl,
	writeln('Te toca mover a un sabueso'),
	writeln('Numero de sabueso -> 0,1,2,3'),
	read(S),
	writeln('Direccion del sabueso -> 3,4'),
	read(D),
	((numeroSabueso(S),direccionSabueso(D),dameSabueso(S,LS,[X,Y]),posSgte(D,X,Y,NX,NY),posicionVacia(NX,NY,XZ,YZ,LS)) -> 
		(
			moverSabueso(S,D,XZ,YZ,LS,NLS)
		)
		;
		(
			writeln('El movimiento no se puede realizar, elija otro movimiento.'),sleep(1),
			turnoSabueso(XZ,YZ,LS,NLS)
		)
	).

% Ejecuta el turno de los sabuesos (En partida 1vsCPU)	
turnoSabuesoCPUFacil(XZ,YZ,LS,NLS) :- 
	limpiar,
	pintarTablero(XZ,YZ,LS),
	nl,
	(
	% Se ejecutan heurísticas en orden de mejor a peor, si con alguna se ejecuta movimiento entonces hacemos corte
		mueveElMasLejano(XZ,YZ,LS,NLS), !, writeln('mueveElMasLejano');
		mueveAleatorio(XZ,YZ,LS,NLS), !, writeln('mueveAleatorio')
	).

% Ejecuta el turno de los sabuesos (En partida 1vsCPU)	
turnoSabuesoCPUDificil(XZ,YZ,LS,NLS) :- 
	limpiar,
	pintarTablero(XZ,YZ,LS),
	nl,
	(
	% Se ejecutan heurísticas en orden de mejor a peor, si con alguna se ejecuta movimiento entonces hacemos corte
		mueveSiVaAGanar(XZ,YZ,LS,NLS), !, writeln('mueveSiVaAGanar');
		mueveSiHaceBarrera(XZ,YZ,LS,NLS), !, writeln('mueveSiHaceBarrera');
		mueveElMasLejano(XZ,YZ,LS,NLS), !, writeln('mueveElMasLejano');
		mueveMenosInaccesibles(XZ,YZ,LS,NLS), !, writeln('mueveMenosInaccesibles')
	).

%%%%%%%%%%%%%%%
% HEURISTICAS % 
%%%%%%%%%%%%%%%

% Heuristica 1: Si los sabueso ganan, ejecuta el movimiento
mueveSiVaAGanar(XZ,YZ,LS,NLS) :- direccionSabueso(D), numeroSabueso(S), moverSabueso(S,D,XZ,YZ,LS,NLS), ganaSabuesos(XZ,YZ,NLS), !.
% Heuristica 2: Si los sabueso hacen una barrera al zorro, ejecuta el movimiento
mueveSiHaceBarrera(XZ,YZ,LS,NLS) :- direccionSabueso(D), numeroSabueso(S), moverSabueso(S,D,XZ,YZ,LS,NLS), hayBarrera(_,0,NLS), !.	
% Heuristica 3: Intenta mover el sabueso más lejano en dirección hacia el zorro, si no puede, intenta mover los demás del mismo modo
mueveElMasLejano(XZ,YZ,LS,NLS) :- quickSort(LS,Lord), moverHaciaZorro(Lord,XZ,YZ,LS,NLS), !.
% Heuristica 4: Mueve el primer sabueso que sea capaz de moverse
mueveAleatorio(XZ,YZ,LS,NLS) :- numeroSabueso(S), direccionSabueso(D), moverSabueso(S,D,XZ,YZ,LS,NLS), !.
% TODO Testear
% Heuristica 5: Mueve el sabueso más cercano al zorro
mueveMasCercano(XZ,YZ,LS,NLS) :- eligeSabuesoMasCercano(S,D,XZ,YZ,LS), moverSabueso(S,D,XZ,YZ,LS,NLS), !.
% Heurística 6: Mueve el sabueso que menos casillas inaccesibles (para el sabueso) deja
mueveMenosInaccesibles(XZ,YZ,LS,NNLS) :- findall(
		[N,NLS],
		(numeroSabueso(S), direccionSabueso(D), moverSabueso(S,D,XZ,YZ,LS,NLS), casillasNoAccesibles(XZ,YZ,NLS,N)),
		L
		% Devuelve en L una lista de pares [valorDeHeurística,nuevaListaDeSabuesosDespuesDeMover]
	),
	quickSort(L,L1), % Ordena L crecientemente por valor de heurística (número de casillas no accesibles por los sabuesos)
	member([_,NNLS],L1), % Saco el primero como resultado
	!. % Sólo quiero el primer movimiento con menor valor

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FUNCIONES AUXILIARES GENÉRICAS %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Quicksort
quickSort([],[]).
quickSort([[X,Z]|Xs],Y):-
	xMayMen([X,Z],Xs,Li,Ld), quickSort(Li,LI), quickSort(Ld,LD), concatenar(LI,[[X,Z]|LD],Y).
xMayMen([X,Z],L,Men,May) :- xMenores([X,Z],L,Men), xMayores([X,Z],L,May).
xMenores([_,_],[],[]) :- !.
xMenores([X,T],[[Y,Y1]|W],[[Y,Y1]|Z]) :- X>Y,xMenores([X,T],W,Z),!.
xMenores([X,_],[[Y,_]|W],Z) :- xMenores([X,Y],W,Z),!.
xMayores([_,_],[],[]) :- !.
xMayores([X,T],[[Y,Y1]|W],[[Y,Y1]|Z]) :- X=<Y,xMayores([X,T],W,Z),!.
xMayores([X,T],[[_,_]|W],Z) :- xMayores([X,T],W,Z),!.
concatenar([],L,L) :- !.
concatenar([X|M],L,[X|Y]) :- concatenar(M,L,Y).

% Suma los elementos de una lista
sumaLista([],0).
sumaLista([X|Xs],Total) :- sumaLista(Xs, Suma), Total is X + Suma.

% Calcula mínimo de X e Y
minimo(X,Y,Z):- Z#=min(X,Y).

% Comprobar paridad
par(X) :- X mod 2 #= 0 .
impar(X) :- X mod 2 #= 1.

% Contar individuos
contar(P,C) :- findall(1,P,L),length(L,C).

% Valor absoluto de la diferencia
diferenciaAbs(X,Y,Z):- Z#=abs(X-Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FUNCIONES AUXILIARES DEL JUEGO %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Dada una dirección y posición, devuelve la nueva posición a la que se mueve
posSgte(D,X,Y,X1,Y1) :- 
	(D==1, !, X1 is X-1, Y1 is Y-1);
	(D==2, !, X1 is X-1, Y1 is Y+1);
	(D==3, !, X1 is X+1, Y1 is Y-1);
	(D==4, !, X1 is X+1, Y1 is Y+1).

% Va probando a mover cada sabueso hacia el zorro, para ello tiene que estar en una columna diferente a este
% Recorre una lista de coordenadas X,Y de sabueso y va probando con cada uno

% Devolvemos la posicion de cada uno de los sabuesos
damePosicionSabueso(X,[X|_],0).
damePosicionSabueso(X,[_|Xs],Pos) :- damePosicionSabueso(X,Xs,Pos1), Pos is Pos1+1. 
damePosicionSabueso(_,[],_) :- fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FUNCIONES AUXILIARES DE HEURISTICAS %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%---------------
% (Heurística 2)
% Comprueba si hay una barrera de sabuesos a lo ancho en el tablero
% Si casilla(X,Y)= negra lo siguiente tiene que ser sabueso
% Si casilla(X,Y)= sabueso lo siguiente arriba;abajo;medio tiene que ser sabueso o negra
hayBarrera(_,Y,_) :- longitudTablero(Long), Y#>=Long, !.
hayBarrera(X,Y,LS) :- esNegra(X,Y), Y1 is Y+1, hayBarrera(X,Y1,LS), !;
	esSabueso(X,Y,LS), X1 is X-1, Y1 is Y+1, X2 is X+1, ((hayBarrera(X1,Y1,LS), !); (hayBarrera(X2,Y1,LS), !); hayBarrera(X,Y1,LS)).
% Comprueba si una casilla es negra tal como hemos definido el tablero
esNegra(X,Y) :- longitudTablero(Long), X#<Long, X#>=0, Y#<Long, Y#>=0, ((par(X),par(Y)) ; (impar(X),impar(Y))).
% Comprueba si una casilla tiene un sabueso
esSabueso(X,Y,LS):- member([X,Y],LS).
%---------------


%---------------
% (Heurística 3)
moverHaciaZorro([],_,_,_,_) :- fail.
moverHaciaZorro([[X,Y]|Xs],XZ,YZ,LS,NLS) :- damePosicionSabueso([X,Y],LS,S),
	(
		Y<YZ, moverSabueso(S,4,XZ,YZ,LS,NLS), !;
		Y>YZ, moverSabueso(S,3,XZ,YZ,LS,NLS), !;
		moverHaciaZorro(Xs,XZ,YZ,LS,NLS)
	).
%---------------


%---------------
% (Heurística 5)
eligeSabuesoMasCercano(S,D,XZ,YZ,LS) :- 
	candidatos(XZ,YZ,LS,NLS),
	minDist(XZ,YZ,NLS,_,L),
	(damePosicionSabueso(L,LS,S)->
		((puedeMoverSabueso(S,D,XZ,YZ,LS))-> 
		(D=3);(D=4) 
		);
		(!,fail)
	).
% En NLS tenemos solo los sabuesos que pueden acorralar al zorro	X<XZ	
candidatos(XZ,YZ,[[X,Y]|Xs],[[X,Y]|Zs]) :- X =< XZ, candidatos(XZ,YZ,Xs,Zs).
candidatos(XZ,YZ,[[X,_]|Xs],NLS) :- X > XZ, candidatos(XZ,YZ,Xs,NLS). 
candidatos(_,_,[],[]).
% Comprobamos si puede mover el sabueso S en la direccion D
puedeMoverSabueso(S,D,XZ,YZ,LS) :- numeroSabueso(S), direccionSabueso(D), dameSabueso(S,LS,[X,Y]), posSgte(D,X,Y,NX,NY), posicionVacia(NX,NY,XZ,YZ,LS).
% DistanciaMinima entre el zorro y los sabuesos
minDist(_,_,[],1000,[8,8]).
minDist(XZ,YZ,[[X,Y]|Xs],D,[X,Y]) :- distancia(XZ,YZ,[X,Y],D1), minDist(XZ,YZ,Xs,D2,[_,_]), minimo(D1,D2,D),(D1#=<D).
minDist(XZ,YZ,[[X,Y]|Xs],D,[X1,Y1]) :- distancia(XZ,YZ,[X,Y],D1), minDist(XZ,YZ,Xs,D2,[X1,Y1]), minimo(D1,D2,D),(D2#=<D).
% D es la distancia entre el zorro y uno de los sabuesos candidatos
distancia(XZ,YZ,[X,Y],D):- D#=(X-XZ)*(X-XZ) + (Y-YZ)*(Y-YZ).
%---------------


%---------------
% (Heurística 6)
% Calcula las casillas no accesibles por un sabueso en el tablero
casillasNoAccesibles(XZ,YZ,LS,N) :- findall(N, (fila(F), casillasNoAccesiblesFila(F,XZ,YZ,LS,N)),L), sumaLista(L,N).
% Calcula las casillas no accesibles por un sabueso en una fila
casillasNoAccesiblesFila(X,XZ,YZ,LS,N) :- findall([X,Y],(columnasPosibles(X,Y), casillaFueraRango(X,Y,XZ,YZ,LS)),L), length(L,N).
% Calcula si la casilla es accesible por algun sabueso en un número indeterminado de movimientos
casillaEnRango(X,Y,_,_,LS) :- member([XS,YS],LS), Z1 is X-XS, Z1>0, diferenciaAbs(Y,YS,Z2), Z1-Z2#>=0, !.
% Calcula si la casilla es inaccesible por algun sabueso en cualquier número de movimientos
casillaFueraRango(X,Y,XZ,YZ,LS) :- not(casillaEnRango(X,Y,XZ,YZ,LS)), posicionVacia(X,Y,XZ,YZ,LS), !.
%---------------


%%%%%%%%%%%%%%%%%
% OTROS/PRUEBAS %
%%%%%%%%%%%%%%%%%

candidatosCasillasVerdes(X,_,Z,T,XZ,YZ,LS):- longitudTablero(Long),((Z#=X+1;Z#=X+2;Z#=X+3;Z#=X+4;Z#=X+5;Z#=X+6;Z#=X+7),Z#<Long),T#>(-1),T#<Long,columnasPosibles(Z,T),casillaVacia(Z,T,XZ,YZ,LS).

primeraFilaSabuesos(LS,[[X,Y]|Xs],X):-quickSort(LS,[[X,Y]|Xs]).

casillasOcupadasFilaX(X,Y,XZ,YZ,LS):- columnasPosibles(X,Y), not(posicionVacia(X,Y,XZ,YZ,LS)).

casillasVaciasFilaX(X,XZ,YZ,LS,N):-contar(casillasOcupadasFilaX(X,_,XZ,YZ,LS),N1), N #=4-N1.

casillasAntesPrimerSabueso(_,_,LS,N1) :- primeraFilaSabuesos(LS,_,X),N1#=X*4.

casillasFilaPrimerSabueso(XZ,YZ,LS,N1) :- primeraFilaSabuesos(LS,_,X),casillasVaciasFilaX(X,XZ,YZ,LS,N), N1#=N+X*4.
heuristicaTablero(XZ,YZ,[[X,Y]|Xs],N):-	casillasFilaPrimerSabueso(XZ,YZ,[[X,Y]|Xs],N1), 
	contar(candidatosCasillasVerdes(X,Y,_,_,XZ,YZ,[[X,Y]|Xs]),C),
	N#=(N1+C+N2),
	heuristicaTablero(XZ,YZ,Xs,N2).
heuristicaTablero(_,_,[],0).
casillaVacia(Z,T,XZ,YZ,LS):- not(member([Z,T],LS)),!,(Z=\=XZ,!;T=\=YZ).




